#include <malloc.h>
#include <string.h>

#include "array.h"

struct array_t
{
  int nsize;
  int allocated_size;
  int count;
  DESTRUCT_ARRAY destructor;
  void *data;
};

#define ALLOCATE_CHUNK		16


ARRAY
create_array (int nsize, DESTRUCT_ARRAY destructor)
{
  ARRAY p = malloc (sizeof (struct array_t));
  if (p != NULL && nsize > 0)
    {
      p->nsize = nsize;
      p->data = malloc (nsize * ALLOCATE_CHUNK);
      if (p->data != NULL)
	{
	  p->allocated_size = ALLOCATE_CHUNK;
	  p->count = 0;
	  p->destructor = destructor;
	}
      else
	{
	  free (p);
	  p = NULL;
	}
    }
  else
    {
      free (p);
      p = NULL;
    }
  return p;
}

void
destroy_array (ARRAY p)
{
  int i;
  if (p->destructor != NULL)
    {
      unsigned char *pwk = (unsigned char *) (p->data);
      for (i = 0; i < p->count; i++, pwk += p->nsize)
	{
	  p->destructor (pwk);
	}
    }
  free (p->data);
  free (p);
  return;
}

int
add_array_element (ARRAY p, void *element)
{
  int index = p->count;
  if (p->count + 1 > p->allocated_size)
    {
      void *pwk = realloc (p->data,
			   p->nsize * (p->allocated_size + ALLOCATE_CHUNK));
      if (pwk == NULL)
	{
	  return -1;
	}
      p->allocated_size += ALLOCATE_CHUNK;
      p->data = pwk;
    }
  memcpy ((unsigned char *) (p->data) + index * p->nsize, element, p->nsize);
  p->count++;
  return index;
}

void
remove_array_element (ARRAY p, int index)
{
  if (index >= 0 && index < p->count)
    {
      unsigned char *pwk = (unsigned char *) (p->data);
      if (p->destructor != NULL)
	{
	  p->destructor (pwk + index * p->nsize);
	}
      if (index < p->count - 1)
	{
	  memcpy (pwk + index * p->nsize, pwk + (index + 1) * p->nsize,
		  (p->count - index - 1) * p->nsize);
	}
      p->count--;
    }
  return;
}

void
remove_array_all (ARRAY p)
{
  p->count = 0;
  return;
}

void *
get_array_element (ARRAY p, int index)
{
  if (index >= 0 && index < p->count)
    {
      return (unsigned char *) (p->data) + index * p->nsize;
    }
  return NULL;
}

int
get_array_count (ARRAY p)
{
  return p->count;
}

void *
get_array (ARRAY p)
{
  return p->data;
}

int
find_array_element (ARRAY p, void *element, COMPARE_ARRAY compare)
{
  int i;
  unsigned char *pwk = (unsigned char *) (p->data);
  for (i = 0; i < p->count; i++, pwk += p->nsize)
    {
      if (compare (pwk, element) == 0)
	{
	  return i;
	}
    }
  return -1;
}



/* EOF */
