/* CarnageHeart portable team tool */
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "array.h"
#include "chteam.h"

#define LINEBUFFER_SIZE		2048

/* compile switch */
#ifndef USE_EUC			/* use EUC-JP for I/O code */
#define USE_EUC		0
#endif

#if USE_EUC
#include "jkcv.h"
#endif

/* constant value */
static const unsigned char HeaderData[HEADER_SIZE] = {
  0x30, 0x2e, 0x30, 0x2e, 0x31, 0x32
};

static const unsigned char PlayerName[PLAYERNAME_SIZE] = {
  0x83, 0x76, 0x83, 0x8c, 0x83, 0x43, 0x83, 0x84, 0x81, 0x5b, 0x00, 0xff,
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0xff, 0xff, 0xff, 0xff
};

static const char tokenSep[] = "\t\r\n";

static const char *Title = "CarnageHeart portable TEAM converter Version 1.7";

/* System variable */
static int VerboseMode = 0;	/* output verbose message */
static int DumpOke = 0;		/* Dump OKE list instead of Team list */
static int AutoTeamName = 0;	/* Automatic making team-name when build */
static FILE *ErrOut;		/* Error output */

static ARRAY OkeArray = NULL;
static ARRAY TeamArray = NULL;
static ARRAY ChpArray = NULL;
static unsigned char *ploadBuffer = NULL;

static void (*pRepositionFunc) (TEAMBLOCK *, int, char[9]);

/* tiny random generator */
static int
getrand (int nmin, int nmax)
{
  int n = nmax - nmin + 1;
  /* too simple random value. it need more better code.;) */
  int nbase = RAND_MAX / n;
  int nceil = nbase * n;
  int r;
  do
    {
      r = rand ();
    }
  while (r >= nceil);
  return r / nbase + nmin;
}

/* compare functions */
static int
compare_pointer (void *p1, void *p2)
{
  void *pt1 = *((void **) p1);
  void *pt2 = *((void **) p2);
  if (pt1 == pt2)
    {
      return 0;
    }
  else
    {
      return -1;
    }
}

static int
compare_chp (void *p1, void *p2)
{
  CHPBLOCK *pblk1 = *((CHPBLOCK **) p1);
  CHPBLOCK *pblk2 = *((CHPBLOCK **) p2);
  return strcmp (pblk1->filename, pblk2->filename);
}


/* load CHPfile */
static int
loadCHPfile (const char *pfname, unsigned char *pdata)
{
  size_t nread;
  FILE *iptr;
  iptr = fopen (pfname, "rb");
  if (iptr == NULL)
    {
      /* file open error */
      fprintf (ErrOut, "fail to open %s : %s\n", pfname, strerror (errno));
      return -1;
    }
  memset (pdata, 0, CHP_BUFFER_SIZE);
  nread = fread (pdata, 1, CHP_BUFFER_SIZE, iptr);
  if (nread == -1)
    {
      /* occur error */
      fprintf (ErrOut, "fail to read %s : %s\n", pfname, strerror (errno));
      free (pdata);
      return -1;
    }
  fclose (iptr);
  return 0;
}

/* CHPfile data to memory */
static CHPBLOCK *
parseCHPfile (const char *pfname, unsigned char *pdata)
{
  CHPBLOCK *pblk = malloc (sizeof (CHPBLOCK));
  long rpos;
  int i;
  if (pblk == NULL)
    {
      fputs ("not enough memory in parseCHPfile\n", ErrOut);
      return NULL;
    }
  memset (pblk, 0, sizeof (CHPBLOCK));
  strncpy (pblk->filename, pfname, FILENAME_SIZE);
  memcpy (pblk->playername, pdata + CHP_OFS_PLAYERNAME, PLAYERNAME_SIZE);
  rpos = CHP_OFS_OKEDATA;
  for (i = 0; i < OKE_MAX; i++, rpos += OKEDATA_SIZE)
    {
      memcpy (pblk->oke[i].data, pdata + rpos, OKEDATA_SIZE);
      strncpy (pblk->oke[i].filename, pfname, FILENAME_SIZE);
      pblk->oke[i].idx = i;
    }
  rpos = CHP_OFS_TEAMDATA;
  for (i = 0; i < TEAM_MAX; i++, rpos += TEAMDATA_SIZE)
    {
      unsigned char n;
      memcpy (&(pblk->team[i].flag), pdata + rpos + TEAMDATA_OFS_HEADER,
	      sizeof (pblk->team[i].flag));
      memcpy (pblk->team[i].emblem, pdata + rpos + TEAMDATA_OFS_EMBLEM,
	      EMBLEM_SIZE);
      memcpy (pblk->team[i].name, pdata + rpos + TEAMDATA_OFS_NAME,
	      TEAMDATA_NAME_SIZE);
      memcpy (pblk->team[i].owner, pdata + rpos + TEAMDATA_OFS_OWNER,
	      TEAMDATA_OWNER_SIZE);
      n = *(pdata + rpos + TEAMDATA_OFS_OKE1IDX);
      if (n == 0xff)
	{
	  pblk->team[i].poke[0] = NULL;
	  pblk->team[i].pos[0] = 0xff;
	}
      else
	{
	  pblk->team[i].poke[0] = pblk->oke + n;
	  pblk->team[i].pos[0] = *(pdata + rpos + TEAMDATA_OFS_OKE1POS);
	}
      n = *(pdata + rpos + TEAMDATA_OFS_OKE2IDX);
      if (n == 0xff)
	{
	  pblk->team[i].poke[1] = NULL;
	  pblk->team[i].pos[1] = 0xff;
	}
      else
	{
	  pblk->team[i].poke[1] = pblk->oke + n;
	  pblk->team[i].pos[1] = *(pdata + rpos + TEAMDATA_OFS_OKE2POS);
	}
      n = *(pdata + rpos + TEAMDATA_OFS_OKE3IDX);
      if (n == 0xff)
	{
	  pblk->team[i].poke[2] = NULL;
	  pblk->team[i].pos[2] = 0xff;
	}
      else
	{
	  pblk->team[i].poke[2] = pblk->oke + n;
	  pblk->team[i].pos[2] = *(pdata + rpos + TEAMDATA_OFS_OKE3POS);
	}
    }
  return pblk;
}

/* regist team data from CHPfile */
static CHPBLOCK *
registTeam (const char *pfname)
{
  char nameBuffer[NAME_BUFFER_SIZE];
  CHPBLOCK *pblk = NULL;
  if (loadCHPfile (pfname, ploadBuffer) == 0 &&
      (pblk = parseCHPfile (pfname, ploadBuffer)) != NULL)
    {
      void *p;
      int i;
      if (VerboseMode)
	{
	  fprintf (ErrOut, "File:[%s]\n", pfname);
	}
      add_array_element (ChpArray, &pblk);
      for (i = 0; i < TEAM_MAX; i++)
	{
	  int j;
	  if (pblk->team[i].flag == 0 ||
	      (pblk->team[i].poke[0] == NULL &&
	       pblk->team[i].poke[1] == NULL
	       && pblk->team[i].poke[2] == NULL))
	    {
	      /* no team data. try to read next team. */
	      continue;
	    }
	  if (VerboseMode)
	    {
	      strncpy (nameBuffer, (char *) (pblk->team[i].name),
		       TEAMDATA_NAME_SIZE);
	      nameBuffer[TEAMDATA_NAME_SIZE] = '\0';
#if USE_EUC
	      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	      fprintf (ErrOut, "Read team %d : %s", i + 1, nameBuffer);
	      strncpy (nameBuffer, (char *) pblk->team[i].owner,
		       TEAMDATA_OWNER_SIZE);
	      nameBuffer[TEAMDATA_OWNER_SIZE] = '\0';
#if USE_EUC
	      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	      fprintf (ErrOut, "(%s)\n", nameBuffer);
	    }
	  p = &(pblk->team[i]);
	  add_array_element (TeamArray, &p);
	  for (j = 0; j < 3; j++)
	    {
	      if (pblk->team[i].poke[j] != NULL)
		{
		  if (find_array_element (OkeArray, &(pblk->team[i].poke[j]),
					  compare_pointer) == -1)
		    {
		      add_array_element (OkeArray, &(pblk->team[i].poke[j]));
		    }
		}
	    }
	}
    }
  else
    {
      if (VerboseMode)
	{
	  fprintf (ErrOut, "Fail to load %s\n", pfname);
	}
    }
  return pblk;
}



/* dump CHP file */
static void
DumpCHPfile (const char *pfname, FILE * optr)
{
  char nameBuffer[NAME_BUFFER_SIZE];
  CHPBLOCK *pblk = NULL;
  if (loadCHPfile (pfname, ploadBuffer) == 0 &&
      (pblk = parseCHPfile (pfname, ploadBuffer)) != NULL)
    {
      int i;
      if (VerboseMode)
	{
	  fprintf (ErrOut, "File:[%s]\n", pfname);
	}
      for (i = 0; i < TEAM_MAX; i++)
	{
	  int j;
	  if (pblk->team[i].flag == 0 ||
	      (pblk->team[i].poke[0] == NULL &&
	       pblk->team[i].poke[1] == NULL
	       && pblk->team[i].poke[2] == NULL))
	    {
	      /* no team data. try to read next team. */
	      continue;
	    }
	  if (VerboseMode)
	    {
	      fprintf (ErrOut, "Dump team %d\n", i + 1);
	    }
	  if (!DumpOke)
	    {
	      fprintf (optr, "%s\t%d\t", pfname, i);
	    }
	  for (j = 0; j < 3; j++)
	    {
	      if (pblk->team[i].poke[j] == NULL)
		{
		  if (!DumpOke)
		    {
		      fputs ("-1\t-1\t", optr);
		    }
		}
	      else
		{
		  if (DumpOke)
		    {
		      strncpy (nameBuffer,
			       (char *) (pblk->team[i].poke[j]->data) +
			       OKEDATA_OFS_NAME, OKEDATA_NAME_SIZE);
		      nameBuffer[OKEDATA_NAME_SIZE] = '\0';
#if USE_EUC
		      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
		      fprintf (optr, "%s\t%d\t%d\t%s\n", pfname,
			       pblk->team[i].poke[j]->idx,
			       pblk->team[i].pos[j], nameBuffer);
		    }
		  else
		    {
		      fprintf (optr, "%d\t%d\t", pblk->team[i].poke[j]->idx,
			       pblk->team[i].pos[j]);
		    }
		}
	    }
	  if (!DumpOke)
	    {
	      strncpy (nameBuffer, (char *) (pblk->team[i].owner),
		       TEAMDATA_OWNER_SIZE);
	      nameBuffer[TEAMDATA_OWNER_SIZE] = '\0';
#if USE_EUC
	      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	      fprintf (optr, "%s\t", nameBuffer);
	      strncpy (nameBuffer, (char *) (pblk->team[i].name),
		       TEAMDATA_NAME_SIZE);
	      nameBuffer[TEAMDATA_NAME_SIZE] = '\0';
#if USE_EUC
	      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	      fprintf (optr, "%s", nameBuffer);
	      for (j = 0; j < 3; j++)
		{
		  if (pblk->team[i].poke[j] == NULL)
		    {
		      nameBuffer[0] = '\0';
		    }
		  else
		    {
		      strncpy (nameBuffer,
			       (char *) (pblk->team[i].poke[j]->data) +
			       OKEDATA_OFS_NAME, OKEDATA_NAME_SIZE);
		      nameBuffer[OKEDATA_NAME_SIZE] = '\0';
#if USE_EUC
		      convertS2E (nameBuffer, strlen (nameBuffer));
#endif
		    }
		  fprintf (optr, "\t%s", nameBuffer);
		}
	      fputc ('\n', optr);
	    }
	}
      if (VerboseMode)
	{
	  fputs ("Completed\n", ErrOut);
	}
    }
  else
    {
      if (VerboseMode)
	{
	  fprintf (ErrOut, "Fail to dump %s\n", pfname);
	}
    }
  if (pblk != NULL)
    {
      free (pblk);
    }
  return;
}

/* check OKE data in CHPBLOCK */
static int
checkOkeData (CHPBLOCK * pblk, OKEBLOCK * poke)
{
  int i;
  for (i = 0; i < OKE_MAX; i++)
    {
      if (!strcmp (pblk->oke[i].filename, poke->filename) &&
	  pblk->oke[i].idx == poke->idx)
	{
	  /* found oke */
	  return i;
	}
    }
  /* not found */
  return -1;
}

/* write 4 byte data */
static void
write_4byte (unsigned char *p, int n1, int n2, int n3, int n4)
{
  *p++ = (unsigned char) n1;
  *p++ = (unsigned char) n2;
  *p++ = (unsigned char) n3;
  *p++ = (unsigned char) n4;
  return;
}

/* save CHPBLOCK */
static int
saveCHPBLOCK (CHPBLOCK * pblk, FILE * optr)
{
  struct tm nowtm;
  time_t nowsec;
  unsigned long cksum;
  size_t nwrite;
  int npos;
  int i;
  time (&nowsec);
  nowtm = *localtime (&nowsec);
  nowtm.tm_mon++;
  if (VerboseMode)
    {
      fputs ("Save CHP file...\n", ErrOut);
    }
  memcpy (ploadBuffer, HeaderData, HEADER_SIZE);
  /* trasfer oke data */
  for (npos = CHP_OFS_OKEDATA, i = 0; i < OKE_MAX; i++, npos += OKEDATA_SIZE)
    {
      memcpy (ploadBuffer + npos, pblk->oke[i].data, OKEDATA_SIZE);
      pblk->oke[i].idx = i;
    }
  /* transfer player name */
  memcpy (ploadBuffer + CHP_OFS_PLAYERNAME, pblk->playername,
	  PLAYERNAME_SIZE);
  /* transfer team data */
  memset (ploadBuffer + CHP_OFS_TEAMDATA, 0, TEAMDATA_SIZE * TEAM_MAX);
  for (npos = CHP_OFS_TEAMDATA, i = 0; i < TEAM_MAX;
       i++, npos += TEAMDATA_SIZE)
    {
      int valid_team = 0;
      memcpy (ploadBuffer + npos + TEAMDATA_OFS_EMBLEM, pblk->team[i].emblem,
	      EMBLEM_SIZE);
      memcpy (ploadBuffer + npos + TEAMDATA_OFS_NAME, pblk->team[i].name,
	      TEAMDATA_NAME_SIZE);
      memcpy (ploadBuffer + npos + TEAMDATA_OFS_OWNER, pblk->team[i].owner,
	      TEAMDATA_OWNER_SIZE);
      if (pblk->team[i].poke[0] == NULL)
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE1IDX,
		       0xff, 0xff, 0xff, 0xff);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE1POS,
		       0xff, 0xff, 0xff, 0xff);
	}
      else
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE1IDX,
		       pblk->team[i].poke[0]->idx, 0, 0, 0);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE1POS,
		       pblk->team[i].pos[0], 0, 0, 0);
	  valid_team = 1;
	}
      if (pblk->team[i].poke[1] == NULL)
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE2IDX,
		       0xff, 0xff, 0xff, 0xff);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE2POS,
		       0xff, 0xff, 0xff, 0xff);
	}
      else
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE2IDX,
		       pblk->team[i].poke[1]->idx, 0, 0, 0);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE2POS,
		       pblk->team[i].pos[1], 0, 0, 0);
	  valid_team = 1;
	}
      if (pblk->team[i].poke[2] == NULL)
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE3IDX,
		       0xff, 0xff, 0xff, 0xff);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE3POS,
		       0xff, 0xff, 0xff, 0xff);
	}
      else
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE3IDX,
		       pblk->team[i].poke[2]->idx, 0, 0, 0);
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_OKE3POS,
		       pblk->team[i].pos[2], 0, 0, 0);
	  valid_team = 1;
	}
      if (valid_team)
	{
	  write_4byte (ploadBuffer + npos + TEAMDATA_OFS_HEADER, 2, 0, 0, 0);
	  *(ploadBuffer + npos + TEAMDATA_OFS_YEAR) =
	    (unsigned char) (nowtm.tm_year & 0xff);
	  *(ploadBuffer + npos + TEAMDATA_OFS_YEAR + 1) =
	    (unsigned char) (nowtm.tm_year / 256);
	  *(ploadBuffer + npos + TEAMDATA_OFS_MONTH) =
	    (unsigned char) (nowtm.tm_mon & 0xff);
	  *(ploadBuffer + npos + TEAMDATA_OFS_MONTH + 1) =
	    (unsigned char) (nowtm.tm_mon / 256);
	  *(ploadBuffer + npos + TEAMDATA_OFS_DAY) =
	    (unsigned char) (nowtm.tm_mday & 0xff);
	  *(ploadBuffer + npos + TEAMDATA_OFS_DAY + 1) =
	    (unsigned char) (nowtm.tm_mday / 256);
	}
    }
  cksum = 0;
  for (i = 0; i < CHP_SAVE_SIZE - 4; i++)
    {
      cksum += *(ploadBuffer + i);
    }
  if (VerboseMode)
    {
      fprintf (ErrOut, "Checksum:%08lx\n", cksum);
    }
  *(ploadBuffer + CHP_SAVE_SIZE - 4) = (unsigned char) (cksum & 0xff);
  *(ploadBuffer + CHP_SAVE_SIZE - 3) = (unsigned char) (cksum >> 8);
  *(ploadBuffer + CHP_SAVE_SIZE - 2) = (unsigned char) (cksum >> 16);
  *(ploadBuffer + CHP_SAVE_SIZE - 1) = (unsigned char) (cksum >> 24);
  nwrite = fwrite (ploadBuffer, 1, CHP_SAVE_SIZE, optr);
  if (nwrite != CHP_SAVE_SIZE)
    {
      fprintf (ErrOut, "fail to fwrite in saveCHPBLOCK : %s\n",
	       strerror (errno));
      return -1;
    }
  else
    {
      return 0;
    }
}

/* clear OKE data */
static void
clearOkeData (OKEBLOCK * poke, int ncount)
{
  int i;
  for (i = 0; i < ncount; i++, poke++)
    {
      memset (&(poke->data[OKEDATA_OFS_NAME]), 0xcd, OKEDATA_NAME_SIZE);
      memset (&(poke->data[OKEDATA_OFS_CODE]), 0xcd, OKEDATA_CODE_SIZE);
    }
  return;
}


/* Management team data */
static void
ManageTeamData (FILE * iptr, FILE * optr)
{
  char nameBuffer[NAME_BUFFER_SIZE];
  int teamcount = 0;
  int okecount = 0;
  CHPBLOCK *pblk = malloc (sizeof (CHPBLOCK));
  CHPBLOCK *pcmp_wk = malloc (sizeof (CHPBLOCK));
  char *lineBuffer = malloc (LINEBUFFER_SIZE);
  if (pblk == NULL || pcmp_wk == NULL || lineBuffer == NULL)
    {
      fputs ("not enough memory in ManageTeamData\n", ErrOut);
      exit (1);
    }
  memset (pblk, 0, sizeof (CHPBLOCK));
  memcpy (pblk->playername, PlayerName, PLAYERNAME_SIZE);
  while (fgets (lineBuffer, LINEBUFFER_SIZE, iptr) != NULL)
    {
      int i;
      int nwk;
      CHPBLOCK *pblk_wk;
      char *pwk;
      if (lineBuffer[0] == '\n' || lineBuffer[0] == '#' ||
	  lineBuffer[0] == '\0')
	{
	  /* null line or blank line or comment line */
	  continue;
	}
      pwk = strtok (lineBuffer, tokenSep);
      if (pwk == NULL)
	{
	  /* ?? */
	  continue;
	}
      if (teamcount == TEAM_MAX || okecount == OKE_MAX)
	{
	  /* too many team or oke */
	  fputs ("Too many Team or OKE data!\nTask aborted.\n", ErrOut);
	  teamcount = 0;
	  break;
	}
      strncpy (pcmp_wk->filename, pwk, FILENAME_SIZE);
      nwk = find_array_element (ChpArray, &pcmp_wk, compare_chp);
      if (nwk == -1)
	{
	  pblk_wk = registTeam (pwk);
	}
      else
	{
	  pblk_wk = *((CHPBLOCK **) get_array_element (ChpArray, nwk));
	}
      pwk = strtok (NULL, tokenSep);
      if (pwk == NULL)
	{
	  /* ?? */
	  continue;
	}
      nwk = atoi (pwk);		/* Team index */
      if (nwk < 0 || nwk >= TEAM_MAX)
	{
	  /* out of range */
	  continue;
	}
      pblk->team[teamcount] = pblk_wk->team[nwk];
      for (i = 0; i < 3; i++)
	{
	  if (pblk_wk->team[nwk].poke[i] != NULL)
	    {
	      int noke;
	      noke = checkOkeData (pblk, pblk_wk->team[nwk].poke[i]);
	      if (noke == -1)
		{
		  /* need to regist OKE data */
		  if (okecount == OKE_MAX)
		    {
		      /* ?? */
		      fputs ("Too many OKE data!\nTask aborted.\n", ErrOut);
		      teamcount = 0;
		      i = -1;
		      break;
		    }
		  pblk->oke[okecount] = *(pblk_wk->team[nwk].poke[i]);
		  noke = okecount++;
		}
	      pblk->team[teamcount].poke[i] = &(pblk->oke[noke]);
	    }
	}
      if (i == -1)
	{
	  break;
	}
      if (VerboseMode)
	{
	  strncpy (nameBuffer, (char *) (pblk->team[teamcount].name),
		   TEAMDATA_NAME_SIZE);
	  nameBuffer[TEAMDATA_NAME_SIZE] = '\0';
#if USE_EUC
	  convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	  fprintf (ErrOut, "add Team %d:%s", teamcount + 1, nameBuffer);
	  strncpy (nameBuffer, (char *) (pblk->team[teamcount].owner),
		   TEAMDATA_OWNER_SIZE);
	  nameBuffer[TEAMDATA_OWNER_SIZE] = '\0';
#if USE_EUC
	  convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	  fprintf (ErrOut, "(%s)\n", nameBuffer);
	}
      teamcount++;
    }
  if (teamcount)
    {
      /* save CHP file */
      if (okecount < OKE_MAX)
	{
	  clearOkeData (&(pblk->oke[okecount]), OKE_MAX - okecount);
	}
      if (saveCHPBLOCK (pblk, optr))
	{
	  fputs ("task aborted.\n", ErrOut);
	}
      else
	{
	  if (VerboseMode)
	    {
	      fputs ("Completed\n", ErrOut);
	    }
	}
    }
  else
    {
      if (VerboseMode)
	{
	  fputs ("no team data!\n", ErrOut);
	}
    }
  free (lineBuffer);
  free (pcmp_wk);
  free (pblk);
  return;
}


/* read OKE data */
static OKEBLOCK *
readOke (CHPBLOCK * pblk, int nidx)
{
  CHPBLOCK *pblk_wk;
  int nwk;
  nwk = find_array_element (ChpArray, &pblk, compare_chp);
  if (nwk == -1)
    {
      pblk_wk = registTeam (pblk->filename);
    }
  else
    {
      pblk_wk = *((CHPBLOCK **) get_array_element (ChpArray, nwk));
    }
  return &(pblk_wk->oke[nidx]);
}

/* re-position Horizon */
static void
repositionHorizon (TEAMBLOCK * pteam, int nidx, char posmap[9])
{
  /* conflict */
  int j;
  int n;
  int posy = pteam->pos[nidx] / 3 * 3;
  int posx = pteam->pos[nidx] % 3;
  for (n = posx,j = 0; j < 3; j++,n++)
    {
      n %= 3;
      if (posmap[posy + n] == 0)
	{
	  /* found free-space */
	  pteam->pos[nidx] = posy + n;
	  return;
	}
    }
  fputs ("Can't reposition Oke\n", ErrOut);
  exit(1);
  return;
}

/* re-position Vertical */
static void
repositionVertical (TEAMBLOCK * pteam, int nidx, char posmap[9])
{
  /* conflict */
  int j;
  int n;
  int posy = pteam->pos[nidx] / 3;
  int posx = pteam->pos[nidx] % 3;
  for (n = posy,j = 0; j < 3; j++,n++)
    {
      n %= 3;
      if (posmap[posx + n * 3] == 0)
	{
	  /* found free-space */
	  pteam->pos[nidx] = posx + n * 3;
	  return;
	}
    }
  fputs ("Can't reposition Oke\n", ErrOut);
  exit (1);
  return;
}

/* re-position OKE in Team */
static void
repositionOke (TEAMBLOCK * pteam)
{
  int i;
  char posmap[9];
  memset (posmap, 0, sizeof (posmap));
  for (i = 0; i < 3; i++)
    {
      if (pteam->poke[i] != NULL)
	{
	  if (posmap[pteam->pos[i]] == 0)
	    {
	      posmap[pteam->pos[i]] = 1;
	      pteam->ok_plot[i] = 1;
	    }
	}
    }
  for (i = 0; i < 3; i++)
    {
      if (pteam->poke[i] != NULL && pteam->ok_plot[i] == 0)
	{
	  /* conflict */
	  pRepositionFunc (pteam, i, posmap);
	  posmap[pteam->pos[i]] = 1;
	  pteam->ok_plot[i] = 1;
	}
    }
  return;
}

/* make team name from OKEs */
static void
makeTeamName (TEAMBLOCK * pteam)
{
  char nameParts[3][9];
  int i;
  int n1;
  int n2;
  int nlen;
  int use_length = AutoTeamName * 2;
  int npos = 0;
  memset (pteam->name, 0, TEAMDATA_NAME_SIZE);
  for (i = 0; i < 3; i++)
    {
      if (pteam->poke[i] != NULL)
	{
	  strncpy (nameParts[i],
		   (char *) (pteam->poke[i]->data + OKEDATA_OFS_NAME),
		   use_length);
	  nameParts[i][use_length] = '\0';
	}
      else
	{
	  nameParts[i][0] = '\0';
	}
    }
  n1 = getrand (0, 2);
  n2 = getrand (1, 100);
  nlen = strlen (nameParts[n1]);
  if (nlen)
    {
      strncpy ((char *) (pteam->name + npos), nameParts[n1], nlen);
      npos += nlen;
    }
  if (n2 > 50)
    {
      for (i = 0; i < 2; i++)
	{
	  n1 = (n1 + 1) % 3;
	  nlen = strlen (nameParts[n1]);
	  if (nlen)
	    {
	      strncpy ((char *) (pteam->name + npos), nameParts[n1], nlen);
	      npos += nlen;
	    }
	}
    }
  else
    {
      for (i = 0; i < 2; i++)
	{
	  if (n1 == 0)
	    {
	      n1 = 2;
	    }
	  else
	    {
	      n1--;
	    }
	  nlen = strlen (nameParts[n1]);
	  if (nlen)
	    {
	      strncpy ((char *) (pteam->name + npos), nameParts[n1], nlen);
	      npos += nlen;
	    }
	}
    }
  return;
}

/* Build team data */
static void
BuildTeamData (FILE * iptr, FILE * optr)
{
  char nameBuffer[NAME_BUFFER_SIZE];
  int teamcount = 0;
  int okecount = 0;
  CHPBLOCK *pblk = malloc (sizeof (CHPBLOCK));
  CHPBLOCK *pcmp_wk = malloc (sizeof (CHPBLOCK));
  char *lineBuffer = malloc (LINEBUFFER_SIZE);
  if (pblk == NULL || pcmp_wk == NULL || lineBuffer == NULL)
    {
      fputs ("not enough memory in BuildTeamData\n", ErrOut);
      exit (1);
    }
  memset (pblk, 0, sizeof (CHPBLOCK));
  memcpy (pblk->playername, PlayerName, PLAYERNAME_SIZE);
  while (fgets (lineBuffer, LINEBUFFER_SIZE, iptr) != NULL)
    {
      int i;
      int nwk;
      char *pwk;
      int bcopy_emblem;
      if (lineBuffer[0] == '\n' || lineBuffer[0] == '#' ||
	  lineBuffer[0] == '\0')
	{
	  /* null line or blank line or comment line */
	  continue;
	}
      pwk = strtok (lineBuffer, tokenSep);
      if (pwk == NULL)
	{
	  /* ?? */
	  continue;
	}
      if (teamcount == TEAM_MAX || okecount == OKE_MAX)
	{
	  /* too many team or oke */
	  fputs ("Too many Team or OKE data!\nTask aborted.\n", ErrOut);
	  teamcount = 0;
	  break;
	}
      if (*pwk == '\0' || *pwk == '-')
	{
	  memcpy (pblk->team[teamcount].owner, PlayerName,
		  TEAMDATA_OWNER_SIZE);
	}
      else
	{
	  strncpy ((char *) (pblk->team[teamcount].owner), pwk,
		   TEAMDATA_OWNER_SIZE);
#if USE_EUC
	  convertE2S ((char *) (pblk->team[teamcount].owner),
		      TEAMDATA_OWNER_SIZE);
#endif
	}
      pwk = strtok (NULL, tokenSep);
      if (pwk == NULL)
	{
	  /* ?? */
	  continue;
	}
      if (*pwk == '-')
	{
	  memset (pblk->team[teamcount].name, 0, TEAMDATA_NAME_SIZE);
	}
      else
	{
	  strncpy ((char *) (pblk->team[teamcount].name), pwk,
		   TEAMDATA_NAME_SIZE);
#if USE_EUC
	  convertE2S ((char *) (pblk->team[teamcount].name),
		      TEAMDATA_NAME_SIZE);
#endif
	}
      bcopy_emblem = 0;
      for (i = 0; i < 3; i++)
	{
	  OKEBLOCK *poke;
	  pwk = strtok (NULL, tokenSep);
	  if (pwk == NULL)
	    {
	      /* ?? */
	      break;
	    }
	  strncpy (pcmp_wk->filename, pwk, FILENAME_SIZE);
	  pwk = strtok (NULL, tokenSep);
	  if (pwk == NULL)
	    {
	      /* ?? */
	      break;
	    }
	  nwk = atoi (pwk);	/* Oke index */
	  if (pcmp_wk->filename[0] == '\0' || nwk == -1)
	    {
	      /* no oke */
	      pblk->team[teamcount].poke[i] = NULL;
	      pblk->team[teamcount].pos[i] = -1;
	      /* discard oke postition */
	      pwk = strtok (NULL, tokenSep);
	      /* try to next oke */
	      continue;
	    }
	  if (nwk < 0 || nwk >= OKE_MAX)
	    {
	      /* out of range */
	      break;
	    }
	  poke = readOke (pcmp_wk, nwk);
	  pwk = strtok (NULL, tokenSep);
	  if (pwk == NULL)
	    {
	      /* ?? */
	      break;
	    }
	  nwk = atoi (pwk);	/* Oke position */
	  if (poke == NULL || nwk < 0 || nwk > 8)
	    {
	      /* out of range */
	      pblk->team[teamcount].poke[i] = NULL;
	      pblk->team[teamcount].pos[i] = -1;
	      /* try to next oke */
	      continue;
	    }
	  pblk->team[teamcount].pos[i] = nwk;
	  if (bcopy_emblem == 0)
	    {
	      /* copy emblem */
	      bcopy_emblem = 1;
	      memcpy (pblk->team[teamcount].emblem,
		      poke->data + OKEDATA_OFS_EMBLEM, EMBLEM_SIZE);
	    }
	  nwk = checkOkeData (pblk, poke);
	  if (nwk == -1)
	    {
	      /* need to regist OKE data */
	      if (okecount == OKE_MAX)
		{
		  /* ?? */
		  fputs ("Too many OKE data!\nTask aborted.\n", ErrOut);
		  teamcount = 0;
		  i = -1;
		  break;
		}
	      pblk->oke[okecount] = *poke;
	      nwk = okecount++;
	    }
	  pblk->team[teamcount].poke[i] = &(pblk->oke[nwk]);
	}
      if (i == -1)
	{
	  break;
	}
      for (; i < 3; i++)
	{
	  pblk->team[teamcount].poke[i] = NULL;
	  pblk->team[teamcount].pos[i] = -1;
	}
      repositionOke (&(pblk->team[teamcount]));
      if (pblk->team[teamcount].name[0] == '\0' && AutoTeamName != 0)
	{
	  makeTeamName (&(pblk->team[teamcount]));
	}
      if (VerboseMode)
	{
	  strncpy (nameBuffer, (char *) (pblk->team[teamcount].name),
		   TEAMDATA_NAME_SIZE);
	  nameBuffer[TEAMDATA_NAME_SIZE] = '\0';
#if USE_EUC
	  convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	  fprintf (ErrOut, "add Team %d:%s", teamcount + 1, nameBuffer);
	  strncpy (nameBuffer, (char *) (pblk->team[teamcount].owner),
		   TEAMDATA_OWNER_SIZE);
	  nameBuffer[TEAMDATA_OWNER_SIZE] = '\0';
#if USE_EUC
	  convertS2E (nameBuffer, strlen (nameBuffer));
#endif
	  fprintf (ErrOut, "(%s)\n", nameBuffer);
	}
      teamcount++;
    }
  if (teamcount)
    {
      /* save CHP file */
      if (okecount < OKE_MAX)
	{
	  clearOkeData (&(pblk->oke[okecount]), OKE_MAX - okecount);
	}
      if (saveCHPBLOCK (pblk, optr))
	{
	  fputs ("task aborted.\n", ErrOut);
	}
      else
	{
	  if (VerboseMode)
	    {
	      fputs ("Completed\n", ErrOut);
	    }
	}
    }
  else
    {
      if (VerboseMode)
	{
	  fputs ("no team data!\n", ErrOut);
	}
    }
  free (lineBuffer);
  free (pcmp_wk);
  free (pblk);
  return;
}


/* print usage */
static void
printUsage (void)
{
  puts ("Usage:");
  puts ("chteam -v");
  puts ("        Print version information");
  puts ("chteam [-e] [-V] [-o team-script-filename] -l CHPfilename(s)");
  puts ("        Dump Team");
  puts ("chteam [-e] [-V] [-o team-script-filename] -k CHPfilename(s)");
  puts ("        Dump Oke in Team");
  puts ("chteam [-e] [-V] -o CHPfilename -m [team-script-filename]");
  puts ("        Make new CHPfile from team-script-file");
  puts ("chteam [-rv|-rh] [-a|-a3|-a2] [-e] [-V] -o CHPfilename "
	"-b [oke-script-filename]");
  puts ("        Build new CHPfile from oke-script-file");
  puts ("common option(s):");
  puts ("   -V  output verbose message to stderr");
  puts ("   -e  use stdout instead of stderr");
  return;
}


/* main */
int
main (int argc, char *argv[])
{
  time_t seed;
  int i;
  enum task_mode
  {
    TaskDumpCHP,
    TaskManageTeam,
    TaskBuildTeam
  }
  ntask;
  const char *poutname = NULL;
  FILE *iptr = stdin;
  FILE *optr = stdout;
  ErrOut = stderr;
  if (argc == 1)
    {
      printUsage ();
      return 0;
    }
  time (&seed);
  srand ((unsigned) seed);
  pRepositionFunc = repositionHorizon;
  ploadBuffer = malloc (CHP_BUFFER_SIZE);
  if (ploadBuffer == NULL)
    {
      fputs ("not enough memory for ploadBuffer\n", ErrOut);
      return -1;
    }
  OkeArray = create_array (sizeof (OKEBLOCK *), NULL);
  if (OkeArray == NULL)
    {
      fputs ("not enough memory for OkeArray\n", ErrOut);
      return -1;
    }
  TeamArray = create_array (sizeof (TEAMBLOCK *), NULL);
  if (TeamArray == NULL)
    {
      fputs ("not enough memory for TeamArray\n", ErrOut);
      return -1;
    }
  ChpArray = create_array (sizeof (CHPBLOCK *), NULL);
  if (ChpArray == NULL)
    {
      fputs ("not enough memory for ChpArray\n", ErrOut);
      return -1;
    }

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-v"))
	{
	  /* print version */
	  puts (Title);
	  puts ("Copyright (C) 2008-2010 by T.Fukui,"
		"MANJI-TOU development studio");
	  puts ("compile option:");
#if USE_EUC
	  puts ("[USE EUC-JP I/O]");
#endif
	  return 0;
	}
      else if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "-?"))
	{
	  /* print usage */
	  printUsage ();
	  return 0;
	}
      else if (!strcmp (argv[i], "-V"))
	{
	  /* verbose mode */
	  VerboseMode = 1;
	}
      else if (!strcmp (argv[i], "-e"))
	{
	  /* use stdout instead of stderr */
	  ErrOut = stdout;
	}
      else if (!strcmp (argv[i], "-a") || !strcmp (argv[i], "-a4"))
	{
	  /* automatic making team-name when Build team from OKE */
	  /* use OKE name head 4 */
	  AutoTeamName = 4;
	}
      else if (!strcmp (argv[i], "-a3"))
	{
	  /* automatic making team-name when Build team from OKE */
	  /* use OKE name head 3 */
	  AutoTeamName = 3;
	}
      else if (!strcmp (argv[i], "-a2"))
	{
	  /* automatic making team-name when Build team from OKE */
	  /* use OKE name head 2 */
	  AutoTeamName = 2;
	}
      else if (!strcmp (argv[i], "-o"))
	{
	  /* set output filename */
	  i++;
	  poutname = argv[i];
	}
      else if (!strcmp (argv[i], "-l"))
	{
	  /* dump Team */
	  ntask = TaskDumpCHP;
	  DumpOke = 0;
	  i++;
	  break;
	}
      else if (!strcmp (argv[i], "-k"))
	{
	  /* dump Oke */
	  ntask = TaskDumpCHP;
	  DumpOke = 1;
	  i++;
	  break;
	}
      else if (!strcmp (argv[i], "-m"))
	{
	  ntask = TaskManageTeam;
	  i++;
	  if (i < argc)
	    {
	      iptr = fopen (argv[i], "r");
	      if (iptr == NULL)
		{
		  fprintf (ErrOut, "fail to fopen %s in main :%s\n",
			   argv[i], strerror (errno));
		  return -1;
		}
	    }
	}
      else if (!strcmp (argv[i], "-b"))
	{
	  ntask = TaskBuildTeam;
	  i++;
	  if (i < argc)
	    {
	      iptr = fopen (argv[i], "r");
	      if (iptr == NULL)
		{
		  fprintf (ErrOut, "fail to fopen %s in main :%s\n",
			   argv[i], strerror (errno));
		  return -1;
		}
	    }
	}
      else if (!strcmp (argv[i], "-rv"))
	{
	  pRepositionFunc = repositionVertical;
	}
      else if (!strcmp (argv[i], "-rh"))
	{
	  pRepositionFunc = repositionHorizon;
	}
      else
	{
	  /* illegal argument */
	  fprintf (ErrOut, "[%s] is illegal argument. ignored.\n", argv[i]);
	}
    }
  if (VerboseMode)
    {
      fputs (Title, ErrOut);
      fputs ("\nStart task...", ErrOut);
    }
  switch (ntask)
    {
    case TaskDumpCHP:
      if (VerboseMode)
	{
	  fputs ("Dump Team data\n", ErrOut);
	}
      if (poutname != NULL)
	{
	  optr = fopen (poutname, "w");
	  if (optr == NULL)
	    {
	      fprintf (ErrOut, "fail to fopen %s in main : %s\n",
		       poutname, strerror (errno));
	      return -1;
	    }
	}
      for (; i < argc; i++)
	{
	  if (!strcmp (argv[i], "--"))
	    {
	      /* end of filelist */
	      break;
	    }
	  DumpCHPfile (argv[i], optr);
	}
      break;
    case TaskManageTeam:
      if (VerboseMode)
	{
	  fputs ("Management Team data\n", ErrOut);
	}
      if (poutname != NULL)
	{
	  optr = fopen (poutname, "wb");
	  if (optr == NULL)
	    {
	      fprintf (ErrOut, "fail to fopen %s in main : %s\n",
		       poutname, strerror (errno));
	      return -1;
	    }
	}
      ManageTeamData (iptr, optr);
      break;
    case TaskBuildTeam:
      if (VerboseMode)
	{
	  fputs ("Build Team data\n", ErrOut);
	  if (AutoTeamName)
	    {
	      fputs ("[Automatic TeamName Enabled]\n", ErrOut);
	    }
	}
      if (poutname != NULL)
	{
	  optr = fopen (poutname, "wb");
	  if (optr == NULL)
	    {
	      fprintf (ErrOut, "fail to fopen %s in main : %s\n",
		       poutname, strerror (errno));
	      return -1;
	    }
	}
      BuildTeamData (iptr, optr);
      break;
    default:
      fputs ("Illegal task code\n", ErrOut);
      break;
    }
  fclose (optr);
  destroy_array (ChpArray);
  destroy_array (TeamArray);
  destroy_array (OkeArray);
  return 0;
}



/* EOF */
