/* built-in function for KANJI manipulation */
#ifndef JKCV_INLINE_FUNCTION
#define JKCV_INLINE_FUNCTION

/* check Shift-JIS 1st byte */
static int
is1stShiftJis (unsigned char n)
{
  return ((n >= 0x81 && n <= 0x9f) || (n >= 0xe0 && n <= 0xfc)) ? 1 : 0;
}

/* convert Shift-JIS to JIS */
static unsigned short
ShiftJisToJis (unsigned short n)
{
  unsigned char c1 = (unsigned char) (n >> 8);
  unsigned char c2 = (unsigned char) (n & 0xff);
  if (0x7f <= c2)
    {
      c2--;
    }
  if (c1 < 0xe0)
    {
      if (c2 >= 0x9e)
	{
	  c2 -= 0x7d;
	  c1 = (c1 - 0x81) * 2 + 0x22;
	}
      else
	{
	  c2 -= 0x1f;
	  c1 = (c1 - 0x60) + (c1 - 0x81);
	}
    }
  else
    {
      if (c2 >= 0x9e)
	{
	  c2 -= 0x7d;
	  c1 = (c1 - 0xE0) * 2 + 0x60;
	}
      else
	{
	  c2 -= 0x1f;
	  c1 = (c1 - 0xE0) + (c1 - 0x81);
	}
    }
  return ((unsigned short) c1 << 8) | c2;
}

/* convert JIS to Shift-JIS */
static unsigned short
JisToShiftJis (unsigned short n)
{
  unsigned char c1 = (unsigned char) (n >> 8) - 0x21;
  unsigned char c2 = (unsigned char) (n & 0xff);
  if (c1 & 0x01)
    {
      c2 += 0x7e;
    }
  else
    {
      c2 += 0x1f;
    }
  if (c2 >= 0x7f && c2 <= 0x9d)
    {
      c2++;
    }
  c1 >>= 1;
  if (c1 >= 0x1f)
    {
      c1 += 0xc1;
    }
  else
    {
      c1 += 0x81;
    }
  return ((unsigned short) c1 << 8) | c2;
}

/* convert Shift-JIS to EUC-JP */
static const char *
convertS2E (char *p, int nlen)
{
  int i;
  const char *pret = p;
  for (i = 0; i < nlen && *p != '\0'; i++)
    {
      unsigned char c = *p++;
      if (is1stShiftJis (c))
	{
	  /* first byte of KANJI */
	  unsigned char c2 = *p++;
	  unsigned int njis = ShiftJisToJis (((unsigned int) c << 8) | c2);
	  njis |= 0x8080;	/* MBS up */
	  *(p - 2) = (unsigned char) (njis >> 8);
	  *(p - 1) = (unsigned char) njis;
	  i++;
	}
    }
  return pret;
}

/* convert EUC-JP to Shift-JIS */
static const char *
convertE2S (char *p, int nlen)
{
  int i;
  const char *pret = p;
  for (i = 0; i < nlen && *p != '\0'; i++)
    {
      unsigned char c = *p++;
      if (c & 0x80)
	{
	  /* first byte of KANJI */
	  unsigned char c2 = *p++;
	  unsigned int neuc = ((unsigned int) c << 8) | c2;
	  unsigned int nsjis = JisToShiftJis (neuc & 0x7f7f);
	  *(p - 2) = (unsigned char) (nsjis >> 8);
	  *(p - 1) = (unsigned char) nsjis;
	  i++;
	}
    }
  return pret;
}

#endif
/* EOF */
